@@ -1,6 +1,6 @@
 /*----------------------------------------------------------------------------*/
 /*  CP2K: A general program to perform molecular dynamics simulations         */
-/*  Copyright 2000-2025 CP2K developers group <https://cp2k.org>              */
+/*  Copyright 2000-2026 CP2K developers group <https://cp2k.org>              */
 /*                                                                            */
 /*  SPDX-License-Identifier: BSD-3-Clause                                     */
 /*----------------------------------------------------------------------------*/
@@ -10,13 +10,19 @@
 #include "dbm_internal.h"
 #include "dbm_shard.h"
 
+enum dbm_multiply_cpu_options {
+  DBM_MULTIPLY_BLAS_LIBRARY = 1,
+  DBM_MULTIPLY_TASK_REORDER = 2
+};
+
 /*******************************************************************************
  * \brief Internal routine for executing the tasks in given batch on the CPU.
  * \author Ole Schuett
  ******************************************************************************/
-void dbm_multiply_cpu_process_batch(
-    const int ntasks, const dbm_task_t batch[ntasks], const double alpha,
-    const dbm_pack_t *pack_a, const dbm_pack_t *pack_b, dbm_shard_t *shard_c);
+void dbm_multiply_cpu_process_batch(int ntasks, const dbm_task_t batch[ntasks],
+                                    double alpha, const dbm_pack_t *pack_a,
+                                    const dbm_pack_t *pack_b,
+                                    dbm_shard_t *shard_c, int options);
 
 #endif
 
