@@ -1,6 +1,6 @@
 /*----------------------------------------------------------------------------*/
 /*  CP2K: A general program to perform molecular dynamics simulations         */
-/*  Copyright 2000-2025 CP2K developers group <https://cp2k.org>              */
+/*  Copyright 2000-2026 CP2K developers group <https://cp2k.org>              */
 /*                                                                            */
 /*  SPDX-License-Identifier: BSD-3-Clause                                     */
 /*----------------------------------------------------------------------------*/
@@ -181,7 +181,7 @@ __constant__ double binomial_coef[19][19];
  * \brief Initializes the device's constant memory.
  * \author Ole Schuett
  ******************************************************************************/
-static void init_constant_memory() {
+static void init_constant_memory(void) {
   static bool initialized = false;
   if (initialized) {
     return; // constant memory has to be initialized only once
